/** @file config.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ config ]
*/

#ifdef _UNICODE
#define SI_CONVERT_GENERIC	// use generic simple.ini on all platform
#endif
#include "common.h"
#include "SimpleIni.h"
#include "config.h"
#include "fileio.h"
#include "emu.h"
#include "clocale.h"

extern EMU *emu;

config_t config;
static _TCHAR ini_file[_MAX_PATH];

static bool load_cfg_file();
static void save_cfg_file();
static bool load_ini_file(_TCHAR *file);
static void save_ini_file(_TCHAR *file);
static const _TCHAR *conv_to_npath(const _TCHAR *path);
static const _TCHAR *conv_from_npath(const _TCHAR *npath);

//
// definition for ini file
//
static CSimpleIni *ini = NULL;
#define SECTION_MAIN	_T("")
#define SECTION_CONTROL	_T("control")
#define SECTION_TAPE	_T("tape")
#define SECTION_FDD		_T("fdd")
#define SECTION_SCREEN	_T("screen")
#define SECTION_SOUND	_T("sound")
#define SECTION_PRINTER	_T("printer")
#define SECTION_STATE	_T("state")
#define SECTION_AUTOKEY	_T("autokey")
#define SECTION_MSGBOARD _T("messageboard")
#define SECTION_MENU	_T("menu")
#define SECTION_ROM     _T("rom")
#define SECTION_COMM    _T("comm")
#define SECTION_SNAPSHOT _T("snapshot")
#define SECTION_FONT	_T("font")

void init_config()
{
	// initial settings
	memset(&config, 0, sizeof(config_t));

	config.version1 = FILE_VERSION;
	config.version2 = CONFIG_VERSION;

#ifdef USE_FD1
#if defined(USE_FD8) || defined(USE_FD7)
	for (int i=0; i<8; i++) {
#elif defined(USE_FD6) || defined(USE_FD5)
	for (int i=0; i<6; i++) {
#else
	for (int i=0; i<4; i++) {
#endif
		config.updated_recent_disk_path[i] = true;
	}
#endif

#ifdef USE_DATAREC
	config.updated_recent_datarec_path = true;
#endif

	config.window_mode = 0;
	config.stretch_screen = false;
	config.cutout_screen = false;

	config.sound_frequency = 6;	// 48KHz
	config.sound_latency = 1;	// 100msec

	config.cpu_power = 1;

#ifdef USE_DIPSWITCH
	config.dipswitch = DIPSWITCH_DEFAULT;
#endif
#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	config.monitor_type = 0;
#endif
#ifdef _BML3MK5
	config.sync_irq = true;
#endif
#ifdef USE_SCANLINE
	config.scan_line = false;
#endif
#ifdef _BML3MK5
	config.afterimage = 0;
	config.keepimage = 0;
	config.use_opengl = 0;
	config.disptmg_skew = 0;
	config.curdisp_skew = 0;
	config.gl_filter_type = 1;

	config.volume = 50;
	config.fdd_type = 0;
	config.mount_fdd = 0xf;
	config.io_port = IOPORT_MSK_KANJI | IOPORT_MSK_PSG6;

	config.show_dlg = MSK_INSIDELEDBOX | MSK_SHOWMSGBOARD | MSK_SHOWLEDBOX;

	config.fps_no = -1;

	config.disp_device_no = 0;
	config.screen_width = 0;
	config.screen_height = 0;

	_tcscpy(config.msgboard_info_fontname, _T("ipagp.ttf"));
	_tcscpy(config.msgboard_msg_fontname, _T("ipagp.ttf"));
	config.msgboard_info_fontsize = 22;
	config.msgboard_msg_fontsize = 12;
	_tcscpy(config.menu_fontname, _T("ipagp.ttf"));
	config.menu_fontsize = 12;

	config.realmode_datarec = false;

	config.screen_video_size = 0;

	config.beep_volume = 50;
	config.psg6_volume = 50;
	config.psg9_volume = 50;
	config.relay_volume = 70;
	config.cmt_volume = 30;
	config.fdd_volume = 50;
	config.mute = false;
	config.psg6_mute = false;
	config.psg9_mute = false;
	config.relay_mute = false;
	config.cmt_mute = false;
	config.fdd_mute = false;

	for(int i=0; i<MAX_PRINTER; i++) {
		_tcscpy(config.printer_server_host[i], _T("localhost"));
		config.printer_server_port[i] = 10200;
		config.printer_direct[i] = false;
	}

	for(int i=0; i<MAX_COMM; i++) {
		config.comm_dipswitch[i] = 1;
		_tcscpy(config.comm_server_host[i], _T("localhost"));
		config.comm_server_port[i] = 6890 + i;
		config.comm_server[i] = false;
		config.comm_connect[i] = false;
	}
	config.led_pos = 0;

	config.reckey_recording = false;
	config.reckey_playing = false;

	config.wav_reverse = 0;
	config.wav_half = 1;
	config.wav_correct = 0;
	config.wav_correct_type = 0;

	config.wav_sample_rate = 3;
	config.wav_sample_bits = 0;
#endif
#ifdef _HC80
	config.device_type = 2;	// Nonintelligent ram disk
#endif
#ifdef _PC8801MA
	config.boot_mode = 2;	// V2 mode, 4MHz
	config.cpu_clock_low = true;
#endif
}

void load_config(const _TCHAR *file)
{
	// initial settings
	init_config();

	_tcsncpy(ini_file, file, _MAX_PATH - 1);

	//
	if (load_ini_file(ini_file) != true) {
		load_cfg_file();
	}
}

void save_config()
{
	//
	save_ini_file(ini_file);
	save_cfg_file();
}

void release_config()
{
	delete ini;
	ini = NULL;
}

/// @attention config file is no longer supported.
bool load_cfg_file()
{
	return true;
}

/// @attention config file is no longer supported.
void save_cfg_file()
{
}

bool load_ini_file(_TCHAR *ini_file)
{
	_TCHAR section[100];
	_TCHAR key[100];
	_TCHAR base_path[_MAX_PATH];
	_TCHAR base_name[_MAX_PATH];
	_TCHAR buf[_MAX_PATH];
	long valuel;
	bool valueb;

	if (ini != NULL) {
		return true;
	}

	ini = new CSimpleIni();
#ifdef _UNICODE
	ini->SetUnicode(true);
#endif
	UTILITY::get_dir_and_basename(ini_file, base_path, base_name);

	// load ini file
	SI_Error rc;
#ifdef _WIN32
	if ((rc = ini->LoadFile(ini_file)) != SI_OK) {
#else
	char cini_file[_MAX_PATH];
	UTILITY::cconv_from_native_path(ini_file, cini_file, _MAX_PATH);
	if ((rc = ini->LoadFile(cini_file)) != SI_OK) {
#endif
		emu->out_logf(LOG_WARN,(CT)_("%s couldn't be loaded."), base_name);
		return false;
	}

	// check config file version
	if (_tcscmp(ini->GetValue(SECTION_MAIN, _T("Name"), _T("")), _T(CLASS_NAME)) != 0) {
		emu->out_logf(LOG_WARN,(CT)_("%s is invalid file."), base_name);
		return false;
	}

	config.version1 = (int)ini->GetLongValue(SECTION_MAIN, _T("Version1"));
	if(config.version1 != FILE_VERSION) {
		emu->out_logf(LOG_WARN,(CT)_("%s is invalid version."), base_name);
		init_config();
		return false;
	}
	config.version2 = (int)ini->GetLongValue(SECTION_MAIN, _T("Version2"));
	if(config.version2 != CONFIG_VERSION) {
		emu->out_logf(LOG_INFO,(CT)_("%s is old version."), base_name);
	}

	valuel = ini->GetLongValue(SECTION_CONTROL, _T("CpuPower"), config.cpu_power);
	if (0 <= valuel && valuel <= 5) {
		config.cpu_power = (uint8)valuel;
	}

#ifdef USE_DIPSWITCH
	config.dipswitch = (uint8)(ini->GetLongValue(SECTION_CONTROL, _T("DipSwitch"), config.dipswitch) & 0xff);
#endif

#ifdef USE_FD1
	_tcscpy(buf, config.initial_disk_path);
	_tcsncpy(config.initial_disk_path, conv_to_npath(ini->GetValue(SECTION_FDD, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.initial_disk_path);
	UTILITY::convert_path_separator(config.initial_disk_path);

#if defined(USE_FD8) || defined(USE_FD7)
	for (int i=0; i<8; i++) {
#elif defined(USE_FD6) || defined(USE_FD5)
	for (int i=0; i<6; i++) {
#else
	for (int i=0; i<4; i++) {
#endif
		if (config.version2 < 8) {
			// old version is start base 1
			UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_FDD, i+1);
		} else {
			UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_FDD, i);
		}
		for (int j=0; j<MAX_HISTORY; j++) {
			_stprintf(key, _T("File%d"), j+1);
			_tcscpy(buf, config.recent_disk_path[i][j].path);
			_tcsncpy(config.recent_disk_path[i][j].path, conv_to_npath(ini->GetValue(section, key, buf)), _MAX_PATH-1);
			get_number_in_path(config.recent_disk_path[i][j].path, &config.recent_disk_path[i][j].num);
			UTILITY::convert_path_separator(config.recent_disk_path[i][j].path);
		}
		config.updated_recent_disk_path[i] = true;

		valueb = ini->GetBoolValue(section, _T("MountWhenStartUp"), (config.mount_fdd & (1 << i)) != 0);
		config.mount_fdd = (valueb ? config.mount_fdd | (1 << i) : config.mount_fdd & ~(1 << i));
	}
#endif
#ifdef USE_DATAREC
	_tcscpy(buf, config.initial_datarec_path);
	_tcsncpy(config.initial_datarec_path, conv_to_npath(ini->GetValue(SECTION_TAPE, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.initial_datarec_path);
	UTILITY::convert_path_separator(config.initial_datarec_path);
	for (int j=0; j<MAX_HISTORY; j++) {
		_stprintf(key, _T("File%d"), j+1);
		_tcscpy(buf, config.recent_datarec_path[j].path);
		_tcsncpy(config.recent_datarec_path[j].path, conv_to_npath(ini->GetValue(SECTION_TAPE, key, buf)), _MAX_PATH-1);
		UTILITY::convert_path_separator(config.recent_datarec_path[j].path);
	}
	config.updated_recent_datarec_path = true;
	config.realmode_datarec = ini->GetBoolValue(SECTION_TAPE, _T("RealMode"), config.realmode_datarec);
#ifdef _BML3MK5
	config.wav_reverse = ini->GetBoolValue(SECTION_TAPE, _T("LoadWavReverse"), config.wav_reverse);
	config.wav_half = ini->GetBoolValue(SECTION_TAPE, _T("LoadWavHalf"), config.wav_half);
	config.wav_correct = ini->GetBoolValue(SECTION_TAPE, _T("LoadWavCorrect"), config.wav_correct);
	valuel = ini->GetLongValue(SECTION_TAPE, _T("LoadWavCorrectType"), config.wav_correct_type);
	if (0 <= valuel && valuel <= 1) {
		config.wav_correct_type = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_TAPE, _T("SaveWavSampleRate"), config.wav_sample_rate);
	if (0 <= valuel && valuel <= 3) {
		config.wav_sample_rate = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_TAPE, _T("SaveWavSampleBits"), config.wav_sample_bits);
	if (0 <= valuel && valuel <= 1) {
		config.wav_sample_bits = (uint8)valuel;
	}
#endif
#endif

	config.window_mode = (int)ini->GetLongValue(SECTION_SCREEN, _T("WindowMode"), config.window_mode);
#ifdef _BML3MK5
	config.disp_device_no = (int)ini->GetLongValue(SECTION_SCREEN, _T("DisplayDevice"), config.disp_device_no);
	config.screen_width = (int)ini->GetLongValue(SECTION_SCREEN, _T("ScreenWidth"), config.screen_width);
	config.screen_height = (int)ini->GetLongValue(SECTION_SCREEN, _T("ScreenHeight"), config.screen_height);
#endif
	config.stretch_screen = ini->GetBoolValue(SECTION_SCREEN, _T("StretchScreen"), config.stretch_screen);
	config.cutout_screen = ini->GetBoolValue(SECTION_SCREEN, _T("CutoutScreen"), config.cutout_screen);

	config.sound_frequency = (int)ini->GetLongValue(SECTION_SOUND,  _T("FrequencyNo"), config.sound_frequency);
	config.sound_latency = (int)ini->GetLongValue(SECTION_SOUND,  _T("LatencyNo"), config.sound_latency);

#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	config.monitor_type = ini->GetLongValue(SECTION_SCREEN,  _T("MonitorType"), config.monitor_type);
#endif
#ifdef USE_SCANLINE
	config.scan_line = ini->GetBoolValue(SECTION_SCREEN, _T("ScanLine"), config.scan_line);
#endif
#ifdef _BML3MK5
	config.sync_irq = ini->GetBoolValue(SECTION_CONTROL, _T("SyncIRQ"), config.sync_irq);

	valuel = ini->GetLongValue(SECTION_SCREEN, _T("AfterImage"), config.afterimage);
	if (0 <= valuel && valuel <= 2) {
		config.afterimage = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("KeepImage"), config.keepimage);
	if (0 <= valuel && valuel <= 2) {
		config.keepimage = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("UseOpenGL"), config.use_opengl);
	if (0 <= valuel && valuel <= 2) {
		config.use_opengl = (uint8)valuel;
	}
	ini->Delete(SECTION_SCREEN, _T("UseDirect3D"));	// not use in this app
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("DisptmgSkew"), config.disptmg_skew);
	if (0 <= valuel && valuel <= 2) {
		config.disptmg_skew = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("CurdispSkew"), config.curdisp_skew);
	if (0 <= valuel && valuel <= 2) {
		config.curdisp_skew = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("VideoSize"), config.screen_video_size);
	if (0 <= valuel && valuel <= 1) {
		config.screen_video_size = (uint8)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SCREEN, _T("GLFilterType"), config.gl_filter_type);
	if (0 <= valuel && valuel <= 1) {
		config.gl_filter_type = (uint8)valuel;
	}
	ini->Delete(SECTION_SCREEN, _T("D3DFilterType"));	// not use in this app

	valuel = ini->GetLongValue(SECTION_SOUND,  _T("Volume"), config.volume);
	if (0 <= valuel && valuel <= 100) {
		config.volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_CONTROL, _T("FddType"), config.fdd_type);
	if (0 <= valuel && valuel <= 2) {
		config.fdd_type = (int)valuel;
	}
	config.io_port = ini->GetLongValue(SECTION_CONTROL, _T("IoPort"), config.io_port) & IOPORT_MSK_ALL;

	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("ShowLED"), FLG_SHOWLEDBOX ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | MSK_SHOWLEDBOX) : (config.show_dlg & ~MSK_SHOWLEDBOX);
	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("ShowMessage"), FLG_SHOWMSGBOARD ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | MSK_SHOWMSGBOARD) : (config.show_dlg & ~MSK_SHOWMSGBOARD);
	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("UseJoystick"), FLG_USEJOYSTICK ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | MSK_USEJOYSTICK) : (config.show_dlg & ~MSK_USEJOYSTICK);
//	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("InsideLED"), FLG_INSIDELEDBOX ? true : false);
//	config.show_dlg = valueb ? (config.show_dlg | MSK_INSIDELEDBOX) : (config.show_dlg & ~MSK_INSIDELEDBOX);
	valueb = ini->GetBoolValue(SECTION_CONTROL, _T("EnableLightpen"), FLG_USELIGHTPEN ? true : false);
	config.show_dlg = valueb ? (config.show_dlg | MSK_USELIGHTPEN) : (config.show_dlg & ~MSK_USELIGHTPEN);

	valuel = ini->GetLongValue(SECTION_CONTROL, _T("FpsNo"), config.fps_no);
	if (-1 <= valuel && valuel <= 6) {
		config.fps_no = (int)valuel;
	}

	_tcscpy(buf, config.initial_printer_path);
	_tcsncpy(config.initial_printer_path, conv_to_npath(ini->GetValue(SECTION_PRINTER, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.initial_printer_path);
	UTILITY::convert_path_separator(config.initial_printer_path);
	_tcscpy(buf, config.initial_state_path);
	_tcsncpy(config.initial_state_path, conv_to_npath(ini->GetValue(SECTION_STATE, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.initial_state_path);
	UTILITY::convert_path_separator(config.initial_state_path);
#ifdef USE_AUTO_KEY
	_tcscpy(buf, config.initial_autokey_path);
	_tcsncpy(config.initial_autokey_path, conv_to_npath(ini->GetValue(SECTION_AUTOKEY, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.initial_autokey_path);
	UTILITY::convert_path_separator(config.initial_autokey_path);
#endif
	_tcscpy(buf, config.msgboard_info_fontname);
	_tcsncpy(config.msgboard_info_fontname, conv_to_npath(ini->GetValue(SECTION_MSGBOARD, _T("InfoFontName"), buf)), _MAX_PATH-1);
	_tcscpy(buf, config.msgboard_msg_fontname);
	_tcsncpy(config.msgboard_msg_fontname, conv_to_npath(ini->GetValue(SECTION_MSGBOARD, _T("MessageFontName"), buf)), _MAX_PATH-1);
	ini->Delete(SECTION_MSGBOARD, _T("FontName"));	// no longer support

	valuel = ini->GetLongValue(SECTION_MSGBOARD, _T("InfoFontSize"), config.msgboard_info_fontsize);
	if (1 <= valuel && valuel <= 60) {
		config.msgboard_info_fontsize = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_MSGBOARD, _T("MessageFontSize"), config.msgboard_msg_fontsize);
	if (1 <= valuel && valuel <= 60) {
		config.msgboard_msg_fontsize = (int)valuel;
	}
	_tcscpy(buf, config.menu_fontname);
	_tcsncpy(config.menu_fontname, conv_to_npath(ini->GetValue(SECTION_MENU, _T("FontName"), buf)), _MAX_PATH-1);
	valuel = ini->GetLongValue(SECTION_MENU, _T("FontSize"), config.menu_fontsize);
	if (1 <= valuel && valuel <= 60) {
		config.menu_fontsize = (int)valuel;
	}

	_tcscpy(buf, config.rom_path);
	_tcsncpy(config.rom_path, conv_to_npath(ini->GetValue(SECTION_ROM, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.rom_path);
	UTILITY::convert_path_separator(config.rom_path);

	valuel = ini->GetLongValue(SECTION_SOUND, _T("BeepVolume"), config.beep_volume);
	if (0 <= valuel && valuel <= 100) {
		config.beep_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("6VoicePsgVolume"), config.psg6_volume);
	if (0 <= valuel && valuel <= 100) {
		config.psg6_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("9VoicePsgVolume"), config.psg9_volume);
	if (0 <= valuel && valuel <= 100) {
		config.psg9_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("RelayVolume"), config.relay_volume);
	if (0 <= valuel && valuel <= 100) {
		config.relay_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("CmtVolume"), config.cmt_volume);
	if (0 <= valuel && valuel <= 100) {
		config.cmt_volume = (int)valuel;
	}
	valuel = ini->GetLongValue(SECTION_SOUND, _T("FddVolume"), config.cmt_volume);
	if (0 <= valuel && valuel <= 100) {
		config.fdd_volume = (int)valuel;
	}
	config.mute = ini->GetBoolValue(SECTION_SOUND, _T("Mute"), config.mute);
	config.beep_mute = ini->GetBoolValue(SECTION_SOUND, _T("BeepMute"), config.beep_mute);
	config.psg6_mute = ini->GetBoolValue(SECTION_SOUND, _T("6VoicePsgMute"), config.psg6_mute);
	config.psg9_mute = ini->GetBoolValue(SECTION_SOUND, _T("9VoicePsgMute"), config.psg9_mute);
	config.relay_mute = ini->GetBoolValue(SECTION_SOUND, _T("RelayMute"), config.relay_mute);
	config.cmt_mute = ini->GetBoolValue(SECTION_SOUND, _T("CmtMute"), config.cmt_mute);
	config.fdd_mute = ini->GetBoolValue(SECTION_SOUND, _T("FddMute"), config.fdd_mute);

	for (int i=0; i<MAX_PRINTER; i++) {
		UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_PRINTER, i);

		_tcscpy(buf, config.printer_server_host[i]);
		_tcsncpy(config.printer_server_host[i], ini->GetValue(section, _T("ServerHost"), buf), _MAX_PATH-1);
		valuel = ini->GetLongValue(section, _T("ServerPort"), config.printer_server_port[i]);
		if (0 <= valuel && valuel <= 65535) {
			config.printer_server_port[i] = (int)valuel;
		}
	}

	for (int i=0; i<MAX_COMM; i++) {
		UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_COMM, i);
		valuel = ini->GetLongValue(section, _T("DipSwitch"), config.comm_dipswitch[i]);
		if (1 <= valuel && valuel <= 4) {
			config.comm_dipswitch[i] = (int)valuel;
		}
		_tcscpy(buf, config.comm_server_host[i]);
		_tcsncpy(config.comm_server_host[i], ini->GetValue(section, _T("ServerHost"), buf), _MAX_PATH-1);
		valuel = ini->GetLongValue(section, _T("ServerPort"), config.comm_server_port[i]);
		if (0 <= valuel && valuel <= 65535) {
			config.comm_server_port[i] = (int)valuel;
		}
	}
	_tcscpy(buf, config.snapshot_path);
	_tcsncpy(config.snapshot_path, conv_to_npath(ini->GetValue(SECTION_SNAPSHOT, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.snapshot_path);
	UTILITY::convert_path_separator(config.snapshot_path);

	valuel = ini->GetLongValue(SECTION_CONTROL, _T("LEDPosition"), config.led_pos);
	if (0 <= valuel && valuel <= 3) {
		config.led_pos = (uint8)valuel;
	}

	_tcscpy(buf, config.font_path);
	_tcsncpy(config.font_path, conv_to_npath(ini->GetValue(SECTION_FONT, _T("Path"), buf)), _MAX_PATH-1);
	UTILITY::add_path_separator(config.font_path);
	UTILITY::convert_path_separator(config.font_path);

	if (config.fdd_type == 1) {
		// use 3inch fdd
		config.io_port = (config.io_port & ~IOPORT_MSK_FDDALL) | IOPORT_MSK_3FDD;
	} else if (config.fdd_type == 2) {
		// use 5inch fdd
		config.io_port = (config.io_port & ~IOPORT_MSK_FDDALL) | IOPORT_MSK_5FDD;
	}

	// convert from old version data
	if (config.version2 < 8) {
		config.cpu_power++;
	}
#endif
	config.version2 = CONFIG_VERSION;

	emu->out_logf(LOG_INFO,(CT)_("%s was loaded."), base_name);

	return true;
}

void save_ini_file(_TCHAR *ini_file)
{
	_TCHAR comment[100];
	_TCHAR section[100];
	_TCHAR key[100];
	_TCHAR base_path[_MAX_PATH];
	_TCHAR base_name[_MAX_PATH];
	_TCHAR buf[_MAX_PATH];

	if (ini == NULL) {
		return;
	}

	UTILITY::get_dir_and_basename(ini_file, base_path, base_name);

	// section
	UTILITY::stprintf(comment, 100, _T("; %s config file"), _T(DEVICE_NAME));

	// values
	ini->SetValue(SECTION_MAIN, _T("Name"), _T(CLASS_NAME), comment);
	ini->SetValue(SECTION_MAIN, _T("Encording"), _T("UTF-8"));
	ini->SetLongValue(SECTION_MAIN, _T("Version1"), config.version1, NULL, true);
	ini->SetLongValue(SECTION_MAIN, _T("Version2"), config.version2);

	ini->SetLongValue(SECTION_CONTROL, _T("CpuPower"), config.cpu_power);

#ifdef USE_DIPSWITCH
	ini->SetLongValue(SECTION_CONTROL, _T("DipSwitch"), config.dipswitch, NULL, true);
#endif


#ifdef USE_FD1
	ini->SetValue(SECTION_FDD, _T("Path"), conv_from_npath(config.initial_disk_path));

#if defined(USE_FD8) || defined(USE_FD7)
	for (int i=0; i<8; i++) {
#elif defined(USE_FD6) || defined(USE_FD5)
	for (int i=0; i<6; i++) {
#else
	for (int i=0; i<4; i++) {
#endif
		for (int j=0; j<MAX_HISTORY; j++) {
			UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_FDD, i);
			_stprintf(key, _T("File%d"), j+1);
			if (_tcslen(config.recent_disk_path[i][j].path) > 0) {
				_tcscpy(buf, config.recent_disk_path[i][j].path);
				set_number_in_path(buf, config.recent_disk_path[i][j].num);
				ini->SetValue(section, key, conv_from_npath(buf));
			}
		}
		ini->SetBoolValue(section, _T("MountWhenStartUp"), (config.mount_fdd & (1 << i)) != 0);
	}
#endif
#ifdef USE_DATAREC
	ini->SetValue(SECTION_TAPE, _T("Path"), conv_from_npath(config.initial_datarec_path));
	for (int j=0; j<MAX_HISTORY; j++) {
		_stprintf(key, _T("File%d"), j+1);
		if (_tcslen(config.recent_datarec_path[j].path) > 0) {
			ini->SetValue(SECTION_TAPE, key, conv_from_npath(config.recent_datarec_path[j].path));
		}
	}
	ini->SetBoolValue(SECTION_TAPE, _T("RealMode"), config.realmode_datarec);
#ifdef _BML3MK5
	ini->SetBoolValue(SECTION_TAPE, _T("LoadWavReverse"), config.wav_reverse);
	ini->SetBoolValue(SECTION_TAPE, _T("LoadWavHalf"), config.wav_half);
	ini->SetBoolValue(SECTION_TAPE, _T("LoadWavCorrect"), config.wav_correct);
	ini->SetLongValue(SECTION_TAPE, _T("LoadWavCorrectType"), config.wav_correct_type);
	ini->SetLongValue(SECTION_TAPE, _T("SaveWavSampleRate"), config.wav_sample_rate);
	ini->SetLongValue(SECTION_TAPE, _T("SaveWavSampleBits"), config.wav_sample_bits);
#endif
#endif

	ini->SetLongValue(SECTION_SCREEN, _T("WindowMode"), config.window_mode);
#ifdef _BML3MK5
	ini->SetLongValue(SECTION_SCREEN, _T("DisplayDevice"), config.disp_device_no);
	ini->SetLongValue(SECTION_SCREEN, _T("ScreenWidth"), config.screen_width);
	ini->SetLongValue(SECTION_SCREEN, _T("ScreenHeight"), config.screen_height);
#endif
	ini->SetBoolValue(SECTION_SCREEN, _T("StretchScreen"), config.stretch_screen);
	ini->SetBoolValue(SECTION_SCREEN, _T("CutoutScreen"), config.cutout_screen);

	ini->SetLongValue(SECTION_SOUND,  _T("FrequencyNo"), config.sound_frequency);
	ini->SetLongValue(SECTION_SOUND,  _T("LatencyNo"), config.sound_latency);

#if defined(USE_MONITOR_TYPE) || defined(USE_SCREEN_ROTATE)
	ini->SetLongValue(SECTION_SCREEN,  _T("MonitorType"), config.monitor_type);
#endif
#ifdef USE_SCANLINE
	ini->SetBoolValue(SECTION_SCREEN, _T("ScanLine"), config.scan_line);
#endif
#ifdef _BML3MK5
	ini->SetBoolValue(SECTION_CONTROL, _T("SyncIRQ"), config.sync_irq);

	ini->SetLongValue(SECTION_SCREEN, _T("AfterImage"), config.afterimage);
	ini->SetLongValue(SECTION_SCREEN, _T("KeepImage"), config.keepimage);
	ini->SetLongValue(SECTION_SCREEN, _T("UseOpenGL"), config.use_opengl);
	ini->SetLongValue(SECTION_SCREEN, _T("DisptmgSkew"), config.disptmg_skew);
	ini->SetLongValue(SECTION_SCREEN, _T("CurdispSkew"), config.curdisp_skew);
	ini->SetLongValue(SECTION_SCREEN, _T("VideoSize"), config.screen_video_size);
	ini->SetLongValue(SECTION_SCREEN, _T("GLFilterType"), config.gl_filter_type);

	ini->SetLongValue(SECTION_SOUND,  _T("Volume"), config.volume);
	ini->SetLongValue(SECTION_CONTROL, _T("FddType"), config.fdd_type);
	ini->SetLongValue(SECTION_CONTROL, _T("IoPort"), config.io_port, NULL, true);

	ini->SetBoolValue(SECTION_CONTROL, _T("ShowLED"), FLG_SHOWLEDBOX ? true : false);
	ini->SetBoolValue(SECTION_CONTROL, _T("ShowMessage"), FLG_SHOWMSGBOARD ? true : false);
	ini->SetBoolValue(SECTION_CONTROL, _T("UseJoystick"), FLG_USEJOYSTICK ? true : false);
//	ini->SetBoolValue(SECTION_CONTROL, _T("InsideLED"), FLG_INSIDELEDBOX ? true : false);
	ini->SetBoolValue(SECTION_CONTROL, _T("EnableLightpen"), FLG_USELIGHTPEN ? true : false);

	ini->SetLongValue(SECTION_CONTROL, _T("FpsNo"), config.fps_no);

	ini->SetValue(SECTION_PRINTER, _T("Path"), conv_from_npath(config.initial_printer_path));
	ini->SetValue(SECTION_STATE, _T("Path"), conv_from_npath(config.initial_state_path));
#ifdef USE_AUTO_KEY
	ini->SetValue(SECTION_AUTOKEY, _T("Path"), conv_from_npath(config.initial_autokey_path));
#endif
	ini->SetValue(SECTION_MSGBOARD, _T("InfoFontName"), conv_from_npath(config.msgboard_info_fontname));
	ini->SetValue(SECTION_MSGBOARD, _T("MessageFontName"), conv_from_npath(config.msgboard_msg_fontname));
	ini->SetLongValue(SECTION_MSGBOARD, _T("InfoFontSize"), config.msgboard_info_fontsize);
	ini->SetLongValue(SECTION_MSGBOARD, _T("MessageFontSize"), config.msgboard_msg_fontsize);
	ini->SetValue(SECTION_MENU, _T("FontName"), conv_from_npath(config.menu_fontname));
	ini->SetLongValue(SECTION_MENU, _T("FontSize"), config.menu_fontsize);
	ini->SetValue(SECTION_ROM, _T("Path"), conv_from_npath(config.rom_path));
#endif

	ini->SetLongValue(SECTION_SOUND, _T("BeepVolume"), config.beep_volume);
	ini->SetLongValue(SECTION_SOUND, _T("6VoicePsgVolume"), config.psg6_volume);
	ini->SetLongValue(SECTION_SOUND, _T("9VoicePsgVolume"), config.psg9_volume);
	ini->SetLongValue(SECTION_SOUND, _T("RelayVolume"), config.relay_volume);
	ini->SetLongValue(SECTION_SOUND, _T("CmtVolume"), config.cmt_volume);
	ini->SetLongValue(SECTION_SOUND, _T("FddVolume"), config.fdd_volume);
	ini->SetBoolValue(SECTION_SOUND, _T("Mute"), config.mute);
	ini->SetBoolValue(SECTION_SOUND, _T("BeepMute"), config.beep_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("6VoicePsgMute"), config.psg6_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("9VoicePsgMute"), config.psg9_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("RelayMute"), config.relay_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("CmtMute"), config.cmt_mute);
	ini->SetBoolValue(SECTION_SOUND, _T("FddMute"), config.fdd_mute);

	for(int i=0; i<MAX_PRINTER; i++) {
		UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_PRINTER, i);
		ini->SetValue(section, _T("ServerHost"), config.printer_server_host[i]);
		ini->SetLongValue(section, _T("ServerPort"), config.printer_server_port[i]);
	}

	for(int i=0; i<MAX_COMM; i++) {
		UTILITY::stprintf(section, 100, _T("%s%d"), SECTION_COMM, i);
		ini->SetLongValue(section, _T("DipSwitch"), config.comm_dipswitch[i]);
		ini->SetValue(section, _T("ServerHost"), config.comm_server_host[i]);
		ini->SetLongValue(section, _T("ServerPort"), config.comm_server_port[i]);
	}

	ini->SetValue(SECTION_SNAPSHOT, _T("Path"), conv_from_npath(config.snapshot_path));

	ini->SetLongValue(SECTION_CONTROL, _T("LEDPosition"), config.led_pos);

	ini->SetValue(SECTION_FONT, _T("Path"), conv_from_npath(config.font_path));

	// save ini file
#ifdef _WIN32
	if (ini->SaveFile(ini_file) == SI_OK) {
#else
	char cini_file[_MAX_PATH];
	UTILITY::cconv_from_native_path(ini_file, cini_file, _MAX_PATH);
	if (ini->SaveFile(cini_file) == SI_OK) {
#endif
		emu->out_logf(LOG_INFO,(CT)_("%s was saved."), base_name);
	} else {
		emu->out_logf(LOG_WARN,(CT)_("%s couldn't be saved."), base_name);
	}
}

const _TCHAR *conv_to_npath(const _TCHAR *path)
{
#ifdef _WIN32
	static _TCHAR npath[_MAX_PATH];
	if (config.version2 >= 9) {
		UTILITY::conv_to_native_path(path, npath, _MAX_PATH);
		return npath;
	}
#endif
	return path;
}

const _TCHAR *conv_from_npath(const _TCHAR *npath)
{
#ifdef _WIN32
	static _TCHAR path[_MAX_PATH];
	UTILITY::conv_from_native_path(npath, path, _MAX_PATH);
	return path;
#else
	return npath;
#endif
}

bool get_number_in_path(_TCHAR *path, int *number)
{
	int len = (int)_tcslen(path);
	bool match = false;
	bool digit = false;
	int pos = len - 1;
	int num = 0;

	// get the number in the end of path
	// ex: "foo.bar:123" -> ":123"
	for (; pos >= 0 && pos >= len - 4; pos--) {
		if (!_istdigit(path[pos]) && path[pos] != _T(':')) {
			break;
		}
		if (_istdigit(path[pos])) {
			digit = true;
		}
		if (digit && path[pos] == _T(':')) {
			match = true;
			break;
		}
	}
	if (match) {
		_stscanf(&path[pos + 1], _T("%d"), &num);
		for(; pos < len; pos++) path[pos] = _T('\0');
	}
	if (number) *number = num;
	return match;
}

bool set_number_in_path(_TCHAR *path, int number)
{
	// 1 .. 9999
	if (number <= 0 || number > 9999) {
		return false;
	}

	int len = (int)_tcslen(path);
	if (len + 6 >= _MAX_PATH) {
		return false;
	}

	_TCHAR buf[8];

	_stprintf(buf, _T(":%d"), number);
	_tcscat(path, buf);
	return true;
}
