/** @file display.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ display ]
*/

#ifndef _DISPLAY_H_
#define _DISPLAY_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"

#define ADDR_VRAM_START		0x0400
#define ADDR_VRAM_END		0x4400

#define ADDR_IGRAM_START	0xa000
#define ADDR_IGRAM_END		0xa800

#define IGRAM_SIZE			0x800

#define SIG_DISPLAY_LPSTB	1
#define SIG_DISPLAY_VSYNC	2

#define VRAM_BUF_HEIGHT	262
#define VRAM_BUF_WIDTH	128

/**
	@brief Display Monitor
*/
class DISPLAY : public DEVICE
{
private:
	DEVICE* d_crtc;

	uint8* regs;
	uint16 start_addr;
	uint16 cursor_addr, cblink;
	bool disp_cursor;
	int cur_left[32];
	int cur_right[32];
	int cur_top;
	int cur_bottom;

//	int ch;
	int   crtc_addr;				///< ffc6
	int   mode_sel;					///< ffd0
	int   interlace_sel;			///< ffd6
	int   ig_enable;				///< ffe9
	bool  ig_mask;					///< compatible for MB-6890/6891

	int   crtc_chr_clocks;			///< crtc clocks 0:1MHz 1:2MHz
	int   crtc_ppch;				///< crtc dot per char w.40:16 w.80:8

	// crtc reg8
	int   interlace;				///< 0:noninterlace 1:interlace (crtc reg8 bit0)
	int   videomode;				///< 1:videomode(interlace only)
	int   disptmg_skew;				///< display timing skew
	int   curdisp_skew;				///< cursor disp skew

	int   bg_color;					///< bg color (mode-sel bit0-bit2)
	int   hireso;					///< 0:normal 1:hireso (mode-sel bit6)
	int   width;					///< width (crtc reg1)
	int   width_sel;				///< 0:w.40 1:w.80 (mode-sel bit7)

	// vertical
	int top_raster;
	int bottom_raster;
	int vsync_top_raster;
	int vsync_bottom_raster;
	int raster_per_char;			///< raster lines per 1 char (16 generally)
	int now_raster;
	int lines_per_2frames;
	// synchronizable vertical range
	int v_total_min[3];
	int v_total_max[3];

	// horizontal
	int left_col[VRAM_BUF_HEIGHT];			///< display timing area left side
	int right_col[VRAM_BUF_HEIGHT];			///< display timing area right side
	int hsync_left_col[VRAM_BUF_HEIGHT];
	int hsync_right_col[VRAM_BUF_HEIGHT];
	int left_diff;
	// synchronizable horizontal range
	int h_total_min[2];
	int h_total_max[2];

	int sw;

	uint32 Rmask;
	uint32 Gmask;
	uint32 Bmask;
	uint8  Rshift;
	uint8  Gshift;
	uint8  Bshift;
	uint32 Rloss;
	uint32 Gloss;
	uint32 Bloss;

	uint8 font[0x1000];
	uint8 chrfont[2][256][16];
	uint8* vram;
	uint8* color_ram;
	uint8* ig_ram;

	// CRT monitor
	typedef struct st_crt_mon {
		uint8 text;	///< vram
		uint8 attr; ///< color ram
	} t_crt_mon;
	t_crt_mon crt_mon[VRAM_BUF_HEIGHT][VRAM_BUF_WIDTH];
	int crt_mon_col;
	int crt_mon_row;
	int crt_mon_addr;
	int crt_mon_page;
	int crt_mon_raster;
	int crt_mon_cur_col;
	int crt_mon_cur_row;

	int screen_left;
	int screen_right;
	int screen_top;
	int screen_bottom;
	int scrnline1_offset;
	scrntype scrnline1[SCREEN_HEIGHT][SCREEN_WIDTH];
	scrntype scrnline2[SCREEN_WIDTH];
	scrntype scrntop[SCREEN_HEIGHT][SCREEN_WIDTH];
	scrntype palette_pc[8];
	int  ppd[8];	// pixel per 1dot
	int  ppch;		// pixel per 1char
	uint8 dof[8];	// dot mask

	bool scanline;
	int  afterimage;
	int  keepimage;

	int skip_frame_count;
	int *crtc_vt_total;		///< crtc total rows (dot unit)
	int *crtc_vt_count;		///< crtc now line
	int *crtc_vt_disp;		///< crtc can display rows (disptmg)

	//for resume
#pragma pack(1)
	struct vm_state_st {
		int   crtc_addr;
		int   ig_enable;
		uint8 ig_mask;

		uint8 chr_clocks;

		char  reserved[6];
	};
#pragma pack()

	bool font_rom_loaded;
	bool font_rom_loaded_at_first;

	void load_font_rom_file();
	uint8 get_font_data(int, int, int);
	void update_cursor();

	void draw_screen_sub();
	void draw_screen_normal(uint16 addr);
public:
	DISPLAY(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier)
	{
		set_class_name("DISPLAY");
	}
	~DISPLAY() {}

	// common functions
	void initialize();
	void reset();
	void update_config();
	void write_io8(uint32 addr, uint32 data);
	void write_signal(int id, uint32 data, uint32 mask);
	void event_frame();
	void update_vram(int v, int clock);

	// unique function
	void set_context_crtc(DEVICE* device) {
		d_crtc = device;
	}
	void set_vram_ptr(uint8* ptr) {
		// vram is main ram!
		vram = ptr;
	}
	void set_color_ram_ptr(uint8* ptr) {
		color_ram = ptr;
	}
	void set_ig_ram_ptr(uint8* ptr) {
		ig_ram = ptr;
	}
	void set_regs_ptr(uint8* ptr) {
		regs = ptr;
	}
	void set_vt_ptr(int* ptrt, int *ptrc, int *ptrd) {
		crtc_vt_total = ptrt;
		crtc_vt_count = ptrc;
		crtc_vt_disp  = ptrd;
	}
	void set_display_size(int left, int top, int right, int bottom);
	void draw_screen();

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);
};

#endif

