/** @file floppy.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.11.05 -

	@brief [ floppy drive ]
*/

#include "floppy.h"
#include "mc6843.h"
#include "mb8866.h"
#include "../disk.h"

#if defined(_DEBUG_FLOPPY) || defined(_DEBUG_MB6843) || defined(_DEBUG_MB8866)
#define OUT_DEBUG emu->out_debugf
#else
#define OUT_DEBUG dummyf
#endif

#define EVENT_IRQ			0
#define EVENT_MOTOR_TIMEOUT	1
#define EVENT_INDEXHOLE_ON	2
#define EVENT_INDEXHOLE_OFF	3

#define EVENT_READY_ON_0	4
#define EVENT_READY_ON_1	5
#define EVENT_READY_ON_2	6
#define EVENT_READY_ON_3	7
#define EVENT_READY_ON_4	8
#define EVENT_READY_ON_5	9
#define EVENT_READY_ON_6	10
#define EVENT_READY_ON_7	11
#define EVENT_MOTOR_OFF		12

// index hole (usually 300rpm -> 200ms)
#define DELAY_INDEX_HOLE	  200000
//#define DELAY_INDEX_HOLE	   50000

// ready on delay time after motor on (us)
#define DELAY_READY_ON		 1000000

// motor warmup time (us)
#define DELAY_MOTOR_WARMUP	  500000
// motor off delay time (us)
#define DELAY_MOTOR_OFF_3	 3000000
#define DELAY_MOTOR_OFF_5S	 5000000
#define DELAY_MOTOR_OFF_5	11000000
#define DELAY_MOTOR_TIMEOUT	60000000

#define DRIVE_MASK	(MAX_DRIVE - 1)


void FLOPPY::cancel_my_event(int event_no) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
}
void FLOPPY::register_my_event(int event_no, int wait) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
	register_event(this, event_no, wait, false, &register_id[event_no]);
}

void FLOPPY::initialize()
{
	// setup/reset floppy drive
	for(int i = 0; i < MAX_DRIVE; i++) {
		disk[i] = new DISK();
	}
	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].side = 0;
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
		fdc[i].ready = 0;
		fdc[i].motor_warmup = 0;
		fdc[i].head_loading = 0;
	}
	ignore_crc = false;

	sidereg = 0;
	sectorcnt = 0;

	head_load = 0;

	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		register_id[i] = -1;
	}

	wav_enable = 0;
	wav_loaded_at_first = false;

	for(int ty = 0; ty < 2; ty++) {
		for(int ft = 0; ft < 2; ft++) {
			wav_data[ft][ty] = NULL;
			wav_size[ft][ty] = 0;
		}
		wav_play_pos[ty] = 0;
		wav_play[ty] = 0;
	}
	_tcscpy(wav_file[0][0], _T("fddseek3.wav"));
	_tcscpy(wav_file[0][1], _T("fddmotor3.wav"));
	_tcscpy(wav_file[1][0], _T("fddseek5.wav"));
	_tcscpy(wav_file[1][1], _T("fddmotor5.wav"));

	register_frame_event(this);
}

void FLOPPY::reset()
{
	load_wav();

	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
		fdc[i].ready = 0;
		fdc[i].motor_warmup = 0;
		fdc[i].head_loading = 0;
	}

	// events were already canceled by EVENT::reset()
	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		register_id[i] = -1;
	}

	drv_num = 0;
	drvsel = 0;

	index_hole = 0;
	density = 0;

	fdd5outreg = 0x81;
	irqflg = false;
	irqflgprev = irqflg;
	drqflg = false;

	ignore_write = false;

	wav_play[0] = 0;
	wav_play[1] = 0;

	if (IOPORT_USE_FDD) {
		register_my_event(EVENT_INDEXHOLE_ON, DELAY_INDEX_HOLE);	// index hole
		register_my_event(EVENT_MOTOR_TIMEOUT, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)
	}
}

void FLOPPY::cancel_my_events()
{
	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		cancel_my_event(i);
	}
}

void FLOPPY::release()
{
	// release d88 handler
	for(int i = 0; i < MAX_DRIVE; i++) {
		delete disk[i];
	}
	for(int i = 0; i < 2; i++) {
		for(int j = 0; j < 2; j++) {
			delete[] wav_data[i][j];
		}
	}
}

void FLOPPY::load_wav()
{
	// load wav file
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];
	wav_header_t	 head;
	wav_fmt_chank_t  fmt;
	wav_data_chank_t data;
	size_t			 data_len;
	size_t           data_max[2];
	PARSEWAV_UTIL    wavu;

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	data_max[0] = (sample_rate >> 3);	// 0.125sec (seek)
	data_max[1] = (sample_rate / 5);	// 0.2sec (moror)
	for (int j=0; j<4; j++) {
		if ((wav_enable & (1 << j)) == 0) {
			wav_data[j >> 1][j & 1] = new uint8[data_max[j & 1]];
			memset(wav_data[j >> 1][j & 1], 128, sizeof(uint8) * data_max[j & 1]);
		}
	}

	FILEIO* fre = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		for (int j=0; j<4; j++) {
			if ((wav_enable & (1 << j)) == 0) {
				UTILITY::stprintf(file_path, _MAX_PATH, _T("%s%s"), app_path, wav_file[j >> 1][j & 1]);
				if(fre->Fopen(file_path, FILEIO_READ_BINARY)) {
					if (wavu.CheckWavFormat(fre, &head, &fmt, &data, &data_len) == 0) {
						wav_size[j >> 1][j & 1] = (int)wavu.ReadWavData(fre, &fmt, data_len, wav_data[j >> 1][j & 1], sample_rate, 8, data_max[j & 1]);
						wav_enable |= (1 << j);
						emu->out_logf(LOG_INFO, (CT)_("%s was loaded."),wav_file[j >> 1][j & 1]);
					}
					fre->Fclose();
				}
			}
		}
	}
	delete fre;

	for (int j=0; j<4; j++) {
		if ((wav_enable & (1 << j)) == 0) {
			wav_size[j >> 1][j & 1] = 0;
			delete[] wav_data[j >> 1][j & 1];
			wav_data[j >> 1][j & 1] = NULL;
			if (!wav_loaded_at_first) {
				OUT_DEBUG((CT)_("%s couldn't be loaded."),wav_file[j >> 1][j & 1]);
			}
		}
	}
	wav_loaded_at_first = true;
}

void FLOPPY::write_io8(uint32 addr, uint32 data)
{
	set_drive_select(data);

//	OUT_DEBUG("fddw a:%04x d:%02x"
//		,addr,data);
}

uint32 FLOPPY::read_io8(uint32 addr)
{
	switch(config.fdd_type) {
		case 1:
			return drvsel;
			break;
		case 2:
			return fdd5outreg;
			break;
	}
	return 0;
}

void FLOPPY::write_signal(int id, uint32 data, uint32 mask)
{
	uint8 nfdd5outreg = fdd5outreg;

	switch (id) {
		case SIG_FLOPPY_ACCESS:
			fdc[drv_num].access = (data & 1) ? true : false;
			break;
		case SIG_FLOPPY_WRITE:
			if(!ignore_write && fdc[drv_num].index < disk[drv_num]->sector_size) {
				disk[drv_num]->sector[fdc[drv_num].index] = (data & mask);
				fdc[drv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITE_TRACK:
			if(!ignore_write && fdc[drv_num].index < disk[drv_num]->track_size) {
				disk[drv_num]->track[fdc[drv_num].index] = (data & mask);
				fdc[drv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITEDELETE:
			if (!ignore_write) {
				disk[drv_num]->deleted = (data & 1) ? 0x10 : 0;
			}
			break;
		case SIG_FLOPPY_STEP:
			{
				int cur_track = fdc[drv_num].track;

				if (data < 0x80) fdc[drv_num].track++;
				else if (data > 0x80) fdc[drv_num].track--;

				if (fdc[drv_num].track < 0) fdc[drv_num].track = 0;
				if (fdc[drv_num].track > 255) fdc[drv_num].track = 255;

				if (cur_track != fdc[drv_num].track) {
					// fddseek sound turn on
					wav_play[0] = 1;
					wav_play_pos[0] = 0;
				}
			}
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			disk[drv_num]->track_size = (data & mask);
			fdc[drv_num].index = 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			fdc[drv_num].track = (data & mask);
			break;
		case SIG_FLOPPY_HEADLOAD:
			head_load = (data & mask);
			fdc[drv_num].head_loading = (head_load ? 6 : 0);
			OUT_DEBUG(_T("fdd %d sig HEADLOAD %x UL:%d"),drv_num,head_load,fdc[drv_num].head_loading);
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			cancel_my_events();
			reset();
			if (now_reset) {
				set_drive_select(0);
			}
			break;
	}

	if (config.fdd_type == 2) {
		// for 5inch mini floppy
		switch (id) {
			case SIG_FLOPPY_IRQ:
				irqflg = (data & mask) ? true : false;
				if (irqflg) {
					nfdd5outreg &= 0xfe;
				} else {
					nfdd5outreg |= 0x01;
				}
				break;
			case SIG_FLOPPY_DRQ:
				drqflg = (data & mask) ? true : false;
				if (drqflg) {
					nfdd5outreg &= 0x7f;
				} else {
					nfdd5outreg |= 0x80;
				}
				break;
		}

//		if (nfdd5outreg != fdd5outreg) {
//			OUT_DEBUG("fdd sig id:%04x d:%02x -> %02x"
//				,id,fdd5outreg,nfdd5outreg);
//		}

		if ((drvsel & 0x40) == 0) {	// nmi non mask
			// nmi interrupt
			switch (id) {
			case SIG_FLOPPY_IRQ:
				if (irqflg != irqflgprev) {
//					set_irq(irqflg);
					register_my_event(EVENT_IRQ, 20);	// delay
					irqflgprev = irqflg;
				}
				break;
	//		case SIG_FLOPPY_DRQ:
	//			set_drq(drqflg);
	//			break;
			}
		}
		fdd5outreg = nfdd5outreg;
	}
}

uint32 FLOPPY::read_signal(int id)
{
	uint32 data = 0;

	switch (id) {
		case SIG_FLOPPY_READ_ID:
			if(fdc[drv_num].index < 6) {
				data = disk[drv_num]->id[fdc[drv_num].index];
				fdc[drv_num].index++;
			}
			break;
		case SIG_FLOPPY_READ:
			if(fdc[drv_num].index < disk[drv_num]->sector_size) {
				data = disk[drv_num]->sector[fdc[drv_num].index];
				fdc[drv_num].index++;
			}
			break;
		case SIG_FLOPPY_READ_TRACK:
			if(fdc[drv_num].index < disk[drv_num]->track_size) {
				data = disk[drv_num]->track[fdc[drv_num].index];
				fdc[drv_num].index++;
			}
			break;
		case SIG_FLOPPY_WRITEPROTECT:
			data = (disk[drv_num]->write_protected || ignore_write) ? 1 : 0;
			break;
		case SIG_FLOPPY_READY:
			data = (disk[drv_num]->inserted && fdc[drv_num].ready >= 2) ? 1 : 0;
			break;
		case SIG_FLOPPY_TRACK0:
			data = (fdc[drv_num].track == 0) ? 1 : 0;
			break;
		case SIG_FLOPPY_INDEX:
			data = index_hole;
			break;
		case SIG_FLOPPY_DELETED:
			data = disk[drv_num]->deleted ? 1 : 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			data = fdc[drv_num].track;
			break;
		case SIG_FLOPPY_SECTOR_NUM:
			data = disk[drv_num]->sector_num;
			break;
		case SIG_FLOPPY_SECTOR_SIZE:
			data = disk[drv_num]->sector_size;
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			data = disk[drv_num]->track_size;
			break;
	}

	return data;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void FLOPPY::event_frame()
{
	for(int i=0; i<MAX_DRIVE; i++) {
		if (fdc[i].head_loading & 0xff) fdc[i].head_loading--;
	}
}

void FLOPPY::event_callback(int event_id, int err)
{
	int event_no = event_id;

	register_id[event_no] = -1;

	switch(event_no) {
		case EVENT_MOTOR_TIMEOUT:
			// motor off
			motor(-1, false);
			break;
		case EVENT_INDEXHOLE_ON:
			// index hole on
			index_hole = 1;
			register_my_event(EVENT_INDEXHOLE_OFF, 300);	// index hole off
			break;
		case EVENT_INDEXHOLE_OFF:
			// index hole off
			index_hole = 0;
			for(int i=0; i<MAX_DRIVE; i++) {
				if (fdc[i].motor_warmup) {
					fdc[i].motor_warmup <<= 1;
					if (fdc[i].motor_warmup > 4) fdc[i].motor_warmup = 0;
					OUT_DEBUG(_T("fdd %d event MOTOR WARMUP %d"), i, fdc[i].motor_warmup);
				}
			}
			register_my_event(EVENT_INDEXHOLE_ON, DELAY_INDEX_HOLE - 300);	// index hole on
			break;
		case EVENT_IRQ:
			cancel_my_event(EVENT_IRQ);
			set_irq(irqflg);
			break;
		case EVENT_READY_ON_0:
		case EVENT_READY_ON_1:
		case EVENT_READY_ON_2:
		case EVENT_READY_ON_3:
		case EVENT_READY_ON_4:
		case EVENT_READY_ON_5:
		case EVENT_READY_ON_6:
		case EVENT_READY_ON_7:
			{
				int i = event_no - EVENT_READY_ON_0;
				fdc[i].ready <<= 1;
				OUT_DEBUG(_T("fdd %d event READY %d"), i, fdc[i].ready);
			}
			break;
		case EVENT_MOTOR_OFF:
			// motor off
			motor(-1, false);
			break;
	}

}

// ----------------------------------------------------------------------------
void FLOPPY::set_drive_select(uint8 data)
{
	uint8 ndrv_num = 0;

	if (now_reset) {
		if (config.fdd_type == 2) {
			data = 0;
		} else if (config.fdd_type == 1) {
			data = 0xff;
		}
	}
	// drive number
	if (config.fdd_type == 2) {
		ndrv_num = (data & DRIVE_MASK);

//		d_fdc5->write_signal(SIG_MB8866_DRIVEREG, ndrv_num, DRIVE_MASK);
	} else if (config.fdd_type == 1) {
		switch(data & 0x0f) {
			case 2:
				ndrv_num = 1;
				break;
			case 4:
				ndrv_num = 2;
				break;
			case 8:
				ndrv_num = 3;
				break;
			default:
				ndrv_num = 0;
				break;
		}
		ndrv_num &= DRIVE_MASK;
		if (drv_num != ndrv_num) {
			// update status
			d_fdc3->write_signal(SIG_MC6843_UPDATESTATUS, 0, 0);
		}
	}
	drv_num = ndrv_num;

	if (config.fdd_type == 2) {
		// 5inch
		// side select
		set_disk_side(drv_num, (data & 0x10) ? 1 : 0);
//		d_fdc5->write_signal(SIG_MB8866_SIDEREG, data, 0x10);
		// density
		density = (data & 0x20) ? 1 : 0;

		// motor
		if ((data & 0x08) == 0 && (drvsel & 0x08) != 0) {
			// motor on -> off
			OUT_DEBUG(_T("fdd %d MOTOR OFF REQUEST data:%02x drvsel:%02x"), drv_num, data, drvsel);
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_OFF_5);	// delay motor off
			data |= 0x08;
			drvsel = data & 0xff;
		} else if ((data & 0x08) != 0) {
			// motor on
			motor(drv_num, true);
			drvsel = data & 0xff;
		} else {
			// motor off
			OUT_DEBUG(_T("fdd %d MOTOR OFF FORCE data:%02x drvsel:%02x"), drv_num, data, drvsel);
			motor(-1, false);
		}
//		d_fdc5->write_signal(SIG_MB8866_MOTOR, drvsel, 0x08);

	} else if (config.fdd_type == 1) {
		// 3inch
		// motor
		if ((data & 0x80) == 0 && (drvsel & 0x80) != 0) {
			// motor on -> off
			OUT_DEBUG(_T("fdd %d MOTOR OFF REQUEST data:%02x drvsel:%02x"), drv_num, data, drvsel);
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_OFF_3);	// delay motor off
			data |= 0x80;
			drvsel = data & 0xff;
		} else if ((data & 0x80) != 0) {
			// motor on
			motor(drv_num, true);
			drvsel = data & 0xff;
		} else {
			// motor off
			OUT_DEBUG(_T("fdd %d MOTOR OFF FORCE data:%02x drvsel:%02x"), drv_num, data, drvsel);
			motor(-1, false);
		}

	}
}

void FLOPPY::motor(int drv, bool val)
{
	if (val) {
		// motor on
		register_my_event(EVENT_MOTOR_TIMEOUT, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)

		// motor sound on
		if (wav_play[1] == 0 && disk[drv]->inserted) {
			wav_play[1] = 1;
		}

		OUT_DEBUG(_T("fdd %d MOTOR ON REQUEST READY:%d WARMUP:%d"), drv, fdc[drv].ready, fdc[drv].motor_warmup);

		// ready on
		if (fdc[drv].ready == 0) {
			fdc[drv].ready = 1;
			register_my_event(EVENT_READY_ON_0 + drv, DELAY_READY_ON);	// delay ready on (1s)
			if (disk[drv]->inserted) {
				fdc[drv].motor_warmup = 1;
			}
		}

		cancel_my_event(EVENT_MOTOR_OFF);

		OUT_DEBUG(_T("fdd %d MOTOR ON  READY:%d WARMUP:%d"), drv, fdc[drv].ready, fdc[drv].motor_warmup);
	} else {
		// motor off
		ignore_write = false;

//		cancel_my_event(EVENT_MOTOR_OFF);

		// motor sound off
		wav_play[1] = 0;
		for(int i=0; i<MAX_DRIVE; i++) {
			if (drv == -1 || drv == i) {
				fdc[i].ready = 0;
				fdc[i].motor_warmup = 0;
			}
			wav_play[1] |= fdc[i].ready;
		}
		if (wav_play[1] < 2) {
			// reset motor flag when all drives stopped
			if (config.fdd_type == 1) {
				// 3inch
				drvsel &= 0x7f;
			} else if (config.fdd_type == 2) {
				// 5inch
				drvsel &= 0xf7;
			}
		}

		OUT_DEBUG(_T("fdd %d MOTOR OFF  SOUND:%d"), drv, wav_play[1]);
	}
}

// ----------------------------------------------------------------------------
// media handler
// ----------------------------------------------------------------------------
bool FLOPPY::search_track()
{
	return disk[drv_num]->get_track(fdc[drv_num].track, fdc[drv_num].side);
}

bool FLOPPY::verify_track()
{
	// verify track number
	for(int i = 0; i < disk[drv_num]->sector_num; i++) {
		if(disk[drv_num]->verify[i] != fdc[drv_num].track) {
			return false;
		}
	}
	return true;
}

int FLOPPY::search_sector()
{
	if (!disk[drv_num]->get_sector(fdc[drv_num].track, fdc[drv_num].side, sectorcnt)) {
		return 1;	// SECTOR NOT FOUND
	}
	sectorcnt++;

	fdc[drv_num].index = 0;

	if (disk[drv_num]->status && !ignore_crc) {
		return 2;	// CRC ERROR
	}

	return 0;
}

int FLOPPY::search_sector(int sect)
{
	return search_sector(sect, false, 0);
}

int FLOPPY::search_sector(int sect, bool compare_side, int side)
{
	int status = 0;
	int sector_num = disk[drv_num]->sector_num;
	sectorcnt = sect - 1;
	if(sectorcnt >= sector_num) {
		sectorcnt = 0;
	}
	int index = sectorcnt - 1;
	for(int i = 0; i < sector_num; i++) {
		// get sector
		index++;
		if(index >= sector_num) {
			index = 0;
		}
		if (!disk[drv_num]->get_sector(fdc[drv_num].track, fdc[drv_num].side, index)) {
			continue;
		}

		// check sector in id field
		if(disk[drv_num]->id[2] != sect) {
			continue;
		}
		// check side in id field
		if(compare_side && disk[drv_num]->id[1] != side) {
			continue;
		}

		// check density
//		if(disk[drv_num]->density) {
//			continue;
//		}

		// sector found
		sectorcnt = index;

		fdc[drv_num].index = 0;

		if (disk[drv_num]->status && !ignore_crc) {
			status |= 2;	// CRC ERROR
		}

		return status;
	}
	return 1;	// SECTOR NOT FOUND
}

bool FLOPPY::make_track()
{
	return disk[drv_num]->make_track(fdc[drv_num].track, fdc[drv_num].side);
}

bool FLOPPY::parse_track()
{
	return disk[drv_num]->parse_track(fdc[drv_num].track, fdc[drv_num].side, density);
}

// ----------------------------------------------------------------------------
// irq / drq
// ----------------------------------------------------------------------------
void FLOPPY::set_irq(bool val)
{
	write_signals(&outputs_irq, val ? 0xffffffff : 0);
}

void FLOPPY::set_drq(bool val)
{
	write_signals(&outputs_drq, val ? 0xffffffff : 0);
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------
bool FLOPPY::open_disk(int drv, const _TCHAR *path, int offset, uint32 flags)
{
	if(drv < MAX_DRIVE) {
		bool rc = disk[drv]->open(path, offset, flags);
		if (config.fdd_type == 2) {
			// 5inch
			if (rc) {
				motor(drv, true);
				register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_OFF_5S);	// delay motor off
			}
		}
		return rc;
	} else {
		return false;
	}
}

void FLOPPY::close_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->close();
		set_disk_side(drv, 0);
		motor(drv, false);
	}
}

int FLOPPY::change_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		set_disk_side(drv, 1 - fdc[drv].side);
		motor(drv, false);
		return fdc[drv].side;
	} else {
		return 0;
	}
}

void FLOPPY::set_disk_side(int drv, int side)
{
	fdc[drv].side = (side & 1) % disk[drv]->num_of_side;
	if (fdc[drv].side) {
		sidereg |= (1 << drv);
	} else {
		sidereg &= ~(1 << drv);
	}
}

int FLOPPY::get_disk_side(int drv)
{
	return (disk[drv]->num_of_side > 1 ? fdc[drv].side % disk[drv]->num_of_side : -1);
}

bool FLOPPY::disk_inserted(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->inserted;
	}
	return false;
}

void FLOPPY::set_drive_type(int drv, uint8 type)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->drive_type = type;
	}
}

uint8 FLOPPY::get_drive_type(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->drive_type;
	}
	return DRIVE_TYPE_UNK;
}

uint8 FLOPPY::fdc_status()
{
	return 0;
}

void FLOPPY::toggle_disk_write_protect(int drv)
{
	if(drv < MAX_DRIVE) {
		if (ignore_write) {
			ignore_write = false;
			disk[drv]->set_write_protect(true);
		}
		disk[drv]->set_write_protect(!(disk[drv]->write_protected));
	}
}

bool FLOPPY::disk_write_protected(int drv)
{
	if(drv < MAX_DRIVE) {
		return (disk[drv]->write_protected || ignore_write);
	}
	return true;
}

// ----------------------------------------------------------------------------
uint16 FLOPPY::get_drive_select()
{
	// for led indicator
	// b3-b0: drive select, b4-b7: 0:green led, 1:red led b8-11:inserted? b12-b13:fdd type
	uint16 data = 0;

	if (config.fdd_type == 2) {
		// 5inch
		for(int i=0; i<MAX_DRIVE; i++) {
			if (fdc[i].motor_warmup || (fdc[i].head_loading > 0 && fdc[i].ready >= 2)) {
				data |= ((1 << i) | (0x10 << i));
			}
		}
	} else if (config.fdd_type == 1) {
		// 3inch
		if (drvsel & 0x80) {
			data = (drvsel & 0x0f) | ((sidereg & 0x0f) << 4);
		}
	}
	// inserted diskette ?
	for(int i=0; i<MAX_DRIVE; i++) {
		if (disk[i]->inserted) {
			data |= (1 << (8 + i));
		}
	}
	// fdd type
	data |= ((config.fdd_type & 3) << 12);

	return data;
}

// ----------------------------------------------------------------------------
void FLOPPY::mix(int32* buffer, int cnt)
{
	int32 *buffer_start = buffer;
	int fdd_type = (config.fdd_type == 2) ? 1 : 0;

	for(int typ=0; typ<2; typ++) {
		if (wav_play[typ] && wav_size[fdd_type][typ]) {
			buffer = buffer_start;
			for(int i=0; i<cnt; i++) {
				*buffer++ += ((int)wav_data[fdd_type][typ][wav_play_pos[typ]] - 128) * max_vol * wav_volume / 12800;
				*buffer++ += ((int)wav_data[fdd_type][typ][wav_play_pos[typ]] - 128) * max_vol * wav_volume / 12800;
				wav_play_pos[typ]++;

				if (wav_play_pos[typ] >= wav_size[fdd_type][typ]) {
					wav_play_pos[typ] = 0;
					if (typ == 0) {
						// seek sound play only one time.
						wav_play[typ] = 0;
						break;
					}
					// motor sound play loop.
//					else {
//					}
				}
			}
		}
	}
}

void FLOPPY::set_volume(int vol)
{
	wav_volume = vol;
}

void FLOPPY::init(int rate, int volume)
{
	sample_rate = rate;
	max_vol = volume;
	wav_volume = 0;

	// load wav file
	load_wav();
}

// ----------------------------------------------------------------------------
void FLOPPY::save_state(FILEIO *fp)
{
	struct vm_state_st_v2 vm_state;

	//
//	vm_state_ident.version = 1;
	vm_state_ident.version = 2;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	for(int i=0; i<22 && i<FLOPPY_MAX_EVENT; i++) {
		vm_state.register_id[i] = register_id[i];
	}
	vm_state.drv_num = drv_num;
	vm_state.drvsel  = drvsel;
	vm_state.sidereg = sidereg;
	vm_state.fdd5outreg = fdd5outreg;
	vm_state.irqflg  = irqflg ? 1 : 0;
	vm_state.drqflg  = drqflg ? 1 : 0;

	fp->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fp->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void FLOPPY::load_state(FILEIO *fp)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st_v1 vm_state_v1;
	struct vm_state_st_v2 vm_state;

	FIND_STATE_CHUNK(fp, vm_state_i);

	// copy values
	if (vm_state_i.version == 1) {
		READ_STATE_CHUNK(fp, vm_state_i, vm_state_v1);

		for(int i=0; i<6; i++) {
			register_id[i] = vm_state_v1.register_id[i];
		}

		drv_num = vm_state_v1.drv_num;
		drvsel  = vm_state_v1.drvsel;
		sidereg = vm_state_v1.sidereg;
		fdd5outreg = vm_state_v1.fdd5outreg;
		irqflg  = vm_state_v1.irqflg ? true : false;
		drqflg  = vm_state_v1.drqflg ? true : false;

		// cannot write to disk when load state from file.
		sectorcnt = 0;
		ignore_write = true;

	} else {
		READ_STATE_CHUNK(fp, vm_state_i, vm_state);

		for(int i=0; i<22 && i<FLOPPY_MAX_EVENT; i++) {
			register_id[i] = vm_state.register_id[i];
		}

		drv_num = vm_state.drv_num;
		drvsel  = vm_state.drvsel;
		sidereg = vm_state.sidereg;
		fdd5outreg = vm_state.fdd5outreg;
		irqflg  = vm_state.irqflg ? true : false;
		drqflg  = vm_state.drqflg ? true : false;

	}
	irqflgprev = irqflg;

	// sound off
	wav_play[0] = 0;
	wav_play[1] = 0;
}
