/** @file sound.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@par Origin
	pcm1bit.h
	@author Sasaji
	@date   2012.01.07 -

	@brief [ sound ]
*/

#ifndef _SOUND_H_
#define _SOUND_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"

#define SOUND_HIGH_QUALITY

#define SIG_SOUND_SIGNAL	0
#define SIG_SOUND_ON		1
#define SIG_SOUND_MUTE		2

#ifdef SOUND_HIGH_QUALITY
#define USE_SOUND_FILTER	3

#define MAX_SAMPLES		1024

#define MAX_OUT_COUNTS	20

#ifdef USE_SOUND_FILTER
#if USE_SOUND_FILTER == 3
#define SAMPLE_RATE_MAX CPU_CLOCKS
#define FILTER_SAMPLES	128
#define FILTER_SAMPLE_NUM	9600
#elif USE_SOUND_FILTER == 2
#define SAMPLE_RATE_MAX 192000
#define FILTER_SAMPLES	32
#define FILTER_SAMPLE_NUM	6400
#define PREV_VOLS_MAX	10
#elif USE_SOUND_FILTER == 1
#define SAMPLE_RATE_MAX 96000
#define SAMPLE_RATE_LCUT 3600
#define SAMPLE_RATE_HCUT 240
#define FILTER_SAMPLES	96
#define FILTER_SAMPLE_NUM	3200
#define PREV_VOLS_MAX	10
#else
#define SAMPLE_RATE_MAX 96000
#define FILTER_SAMPLES	32
#define FILTER_SAMPLE_NUM	3200
#define PREV_VOLS_MAX	16
#endif
#endif

/**
	@brief record turn on/off point of MUSIC_SEL register

	turn on/offʒuL^NX
*/
class SOUND_SAMPLE_POINT
{
private:
	bool   signals[MAX_SAMPLES];
	bool   outs[MAX_SAMPLES];
	uint64 clocks[MAX_SAMPLES];

	bool   stored;
	bool   cur_signal;
	bool   cur_out;
public:
	int    count;
	void add(bool signal, bool out, uint64 clk) {
		if (count >= MAX_SAMPLES) return;
		if (stored) {
			if (signal == cur_signal && out == cur_out) return;
		}

		signals[count] = signal;
		outs[count] = out;
		clocks[count] = clk;
		count++;

		cur_signal = signal;
		cur_out = out;
		stored = true;
	}
	void clear() {
		for(int i=0; i<MAX_SAMPLES; i++) {
			signals[i] = false;
			outs[i] = false;
			clocks[i] = 0;
		}
		count = 0;
		stored = false;
	}
	bool get_signal(int pos) {
		return signals[pos];
	}
	bool get_out(int pos) {
		return outs[pos];
	}
	uint64 get_clock(int pos) {
		return clocks[pos];
	}
	void shift(int cnt) {
		if (cnt > 0) {
			for (int s = 0; s < (count - cnt); s++) {
				signals[s] = signals[s + cnt];
				outs[s] = outs[s + cnt];
				clocks[s] = clocks[s + cnt];
			}
			if (count > cnt) {
				signals[count-cnt] = false;
				outs[count-cnt] = false;
				clocks[count-cnt] = 0;

				count -= cnt;
			} else {
				signals[0] = false;
				outs[0] = false;
				clocks[0] = 0;

				count = 0;
			}
		}
	}
};
#endif	/* SOUND_HIGH_QUALITY */

class FILEIO;

/**
	@brief sound - MUSIC_SEL register
*/
class SOUND : public DEVICE
{
private:
	bool signal, on, mute;
	bool prev_on, prev_mute;

#ifdef SOUND_HIGH_QUALITY
	SOUND_SAMPLE_POINT samples;

	int sample_rate;
//	int sample_clock;
	uint64 cur_clock_s, cur_clock_c;
	uint64 last_registred_clock;
	int event_power;
	int in_vals[FILTER_SAMPLES];
	int widx[MAX_OUT_COUNTS+1];
	int dclk[MAX_OUT_COUNTS+1];
	int wclk;
	int wsample;
	int windex;
	int wresidue;
	int dclocks;
	int dindex;
	int dresidue;
	int prev_event_power;
#if USE_SOUND_FILTER == 3
	int out_vals[FILTER_SAMPLES + 10];
	int prev_vals[FILTER_SAMPLES + 10];
#elif USE_SOUND_FILTER == 2
	int out_vals[FILTER_SAMPLES + 10];
	int prev_vals[FILTER_SAMPLES + 10];
#elif USE_SOUND_FILTER == 1
	double h[FILTER_SAMPLES];
#endif
#endif
	int max_vol;
	int sound_volume;
	int update;
	int prev_vol;

	void get_sample_data(uint64 window_clock, uint64 end_clock);
	void set_widx();

public:
	SOUND(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("SOUND");
	}
	~SOUND() {}

	// common functions
	void initialize();
	void reset();
	void write_signal(int id, uint32 data, uint32 mask);
	void event_frame();
	void mix(int32* buffer, int cnt);
	void set_volume(int volume);

	void update_config();

	void save_state(FILEIO *fp);
	void load_state(FILEIO *fp);

	// unique function
	void init(int rate, int volume);
};

#endif
