/** @file device.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ device base class ]
*/

#ifndef _DEVICE_H_
#define _DEVICE_H_

#include "vm.h"
#include "../emu.h"
#include "../fileio.h"

// max devices connected to the output port
#define MAX_OUTPUT	16

// common signal id
#define SIG_CPU_IRQ	101
#define SIG_CPU_FIRQ	102
#define SIG_CPU_NMI	103
#define SIG_CPU_BUSREQ	104
#define SIG_CPU_DEBUG	201
#ifdef USE_EMU_INHERENT_SPEC
#define SIG_CPU_ADD_DEAD_CYCLE	197
#define SIG_CPU_HALT	198
#define SIG_CPU_RESET	199
#endif

/**
	@brief device base class
*/
class DEVICE
{
protected:
	VM* vm;
	EMU* emu;
public:
	DEVICE(VM* parent_vm, EMU* parent_emu, const char *identifier) : vm(parent_vm), emu(parent_emu)
	{
		prev_device = vm->last_device;
		next_device = NULL;
		if(vm->first_device == NULL) {
			// this is the first device
			vm->first_device = this;
			this_device_id = 0;
		}
		else {
			// this is not the first device
			vm->last_device->next_device = this;
			this_device_id = vm->last_device->this_device_id + 1;
		}
		vm->last_device = this;

		// primary event manager
		event_manager = NULL;

#ifdef USE_EMU_INHERENT_SPEC
		now_reset = false;
#endif
		//
		memset(this_class_name, 0, sizeof(this_class_name));
		memset(this_identifier, 0, sizeof(this_identifier));
		if (identifier != NULL) {
			strncpy(this_identifier, identifier, 4);
		}

		//
		memset(&vm_state_ident, 0, sizeof(vm_state_ident));
		if (identifier != NULL) {
			strncpy(vm_state_ident.identifier, identifier, 4);
		}
		vm_state_ident.device_id = this_device_id;
	}

	virtual ~DEVICE(void) {}

	virtual void initialize() {}
	virtual void release() {}

	virtual void update_config() {}
	virtual void save_state(FILEIO* fio) {}
	virtual void load_state(FILEIO* fio) {}
	virtual void set_class_name(const char *name)
	{
		if (name != NULL) {
			strncpy(this_class_name, name, 12);
			strncpy(vm_state_ident.class_name, name, 12);
		}
	}

	// control
	virtual void reset() {}
	virtual void special_reset() {
		reset();
	}

	// memory bus
	virtual void write_data8(uint32 addr, uint32 data) {}
	virtual uint32 read_data8(uint32 addr) {
		return 0xff;
	}
	virtual void write_data16(uint32 addr, uint32 data) {
		write_data8(addr, data & 0xff);
		write_data8(addr + 1, (data >> 8) & 0xff);
	}
	virtual uint32 read_data16(uint32 addr) {
		uint32 val = read_data8(addr);
		val |= read_data8(addr + 1) << 8;
		return val;
	}
	virtual void write_data32(uint32 addr, uint32 data) {
		write_data8(addr, data & 0xff);
		write_data8(addr + 1, (data >> 8) & 0xff);
		write_data8(addr + 2, (data >> 16) & 0xff);
		write_data8(addr + 3, (data >> 24) & 0xff);
	}
	virtual uint32 read_data32(uint32 addr) {
		uint32 val = read_data8(addr);
		val |= read_data8(addr + 1) << 8;
		val |= read_data8(addr + 2) << 16;
		val |= read_data8(addr + 3) << 24;
		return val;
	}
	virtual void write_data8w(uint32 addr, uint32 data, int* wait) {
		*wait = 0;
		write_data8(addr, data);
	}
	virtual uint32 read_data8w(uint32 addr, int* wait) {
		*wait = 0;
		return read_data8(addr);
	}
	virtual void write_data16w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1;
		write_data8w(addr, data & 0xff, &wait0);
		write_data8w(addr + 1, (data >> 8) & 0xff, &wait1);
		*wait = wait0 + wait1;
	}
	virtual uint32 read_data16w(uint32 addr, int* wait) {
		int wait0, wait1;
		uint32 val = read_data8w(addr, &wait0);
		val |= read_data8w(addr + 1, &wait1) << 8;
		*wait = wait0 + wait1;
		return val;
	}
	virtual void write_data32w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1, wait2, wait3;
		write_data8w(addr, data & 0xff, &wait0);
		write_data8w(addr + 1, (data >> 8) & 0xff, &wait1);
		write_data8w(addr + 2, (data >> 16) & 0xff, &wait2);
		write_data8w(addr + 3, (data >> 24) & 0xff, &wait3);
		*wait = wait0 + wait1 + wait2 + wait3;
	}
	virtual uint32 read_data32w(uint32 addr, int* wait) {
		int wait0, wait1, wait2, wait3;
		uint32 val = read_data8w(addr, &wait0);
		val |= read_data8w(addr + 1, &wait1) << 8;
		val |= read_data8w(addr + 2, &wait2) << 16;
		val |= read_data8w(addr + 3, &wait3) << 24;
		*wait = wait0 + wait1 + wait2 + wait3;
		return val;
	}
	virtual uint32 fetch_op(uint32 addr, int *wait) {
		return read_data8w(addr, wait);
	}
	virtual void latch_address(uint32 addr, int *wait) {
	}
	virtual void write_dma_data8(uint32 addr, uint32 data) {
		write_data8(addr, data);
	}
	virtual uint32 read_dma_data8(uint32 addr) {
		return read_data8(addr);
	}
	virtual void write_dma_data16(uint32 addr, uint32 data) {
		write_data16(addr, data);
	}
	virtual uint32 read_dma_data16(uint32 addr) {
		return read_data16(addr);
	}
	virtual void write_dma_data32(uint32 addr, uint32 data) {
		write_data32(addr, data);
	}
	virtual uint32 read_dma_data32(uint32 addr) {
		return read_data32(addr);
	}

	// i/o bus
	virtual void write_io8(uint32 addr, uint32 data) {}
	virtual uint32 read_io8(uint32 addr) {
#ifdef IOBUS_RETURN_ADDR
		return (addr & 1 ? addr >> 8 : addr) & 0xff;
#else
		return 0xff;
#endif
	}
	virtual void write_io16(uint32 addr, uint32 data) {
		write_io8(addr, data & 0xff);
		write_io8(addr + 1, (data >> 8) & 0xff);
	}
	virtual uint32 read_io16(uint32 addr) {
		uint32 val = read_io8(addr);
		val |= read_io8(addr + 1) << 8;
		return val;
	}
	virtual void write_io32(uint32 addr, uint32 data) {
		write_io8(addr, data & 0xff);
		write_io8(addr + 1, (data >> 8) & 0xff);
		write_io8(addr + 2, (data >> 16) & 0xff);
		write_io8(addr + 3, (data >> 24) & 0xff);
	}
	virtual uint32 read_io32(uint32 addr) {
		uint32 val = read_io8(addr);
		val |= read_io8(addr + 1) << 8;
		val |= read_io8(addr + 2) << 16;
		val |= read_io8(addr + 3) << 24;
		return val;
	}
	virtual void write_io8w(uint32 addr, uint32 data, int* wait) {
		*wait = 0;
		write_io8(addr, data);
	}
	virtual uint32 read_io8w(uint32 addr, int* wait) {
		*wait = 0;
		return read_io8(addr);
	}
	virtual void write_io16w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1;
		write_io8w(addr, data & 0xff, &wait0);
		write_io8w(addr + 1, (data >> 8) & 0xff, &wait1);
		*wait = wait0 + wait1;
	}
	virtual uint32 read_io16w(uint32 addr, int* wait) {
		int wait0, wait1;
		uint32 val = read_io8w(addr, &wait0);
		val |= read_io8w(addr + 1, &wait1) << 8;
		*wait = wait0 + wait1;
		return val;
	}
	virtual void write_io32w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1, wait2, wait3;
		write_io8w(addr, data & 0xff, &wait0);
		write_io8w(addr + 1, (data >> 8) & 0xff, &wait1);
		write_io8w(addr + 2, (data >> 16) & 0xff, &wait2);
		write_io8w(addr + 3, (data >> 24) & 0xff, &wait3);
		*wait = wait0 + wait1 + wait2 + wait3;
	}
	virtual uint32 read_io32w(uint32 addr, int* wait) {
		int wait0, wait1, wait2, wait3;
		uint32 val = read_io8w(addr, &wait0);
		val |= read_io8w(addr + 1, &wait1) << 8;
		val |= read_io8w(addr + 2, &wait2) << 16;
		val |= read_io8w(addr + 3, &wait3) << 24;
		*wait = wait0 + wait1 + wait2 + wait3;
		return val;
	}
	virtual void write_dma_io8(uint32 addr, uint32 data) {
		write_io8(addr, data);
	}
	virtual uint32 read_dma_io8(uint32 addr) {
		return read_io8(addr);
	}
	virtual void write_dma_io16(uint32 addr, uint32 data) {
		write_io16(addr, data);
	}
	virtual uint32 read_dma_io16(uint32 addr) {
		return read_io16(addr);
	}
	virtual void write_dma_io32(uint32 addr, uint32 data) {
		write_io32(addr, data);
	}
	virtual uint32 read_dma_io32(uint32 addr) {
		return read_io32(addr);
	}

	// memory mapped i/o
	virtual void write_memory_mapped_io8(uint32 addr, uint32 data) {
		write_io8(addr, data);
	}
	virtual uint32 read_memory_mapped_io8(uint32 addr) {
		return read_io8(addr);
	}
	virtual void write_memory_mapped_io16(uint32 addr, uint32 data) {
		write_memory_mapped_io8(addr, data & 0xff);
		write_memory_mapped_io8(addr + 1, (data >> 8) & 0xff);
	}
	virtual uint32 read_memory_mapped_io16(uint32 addr) {
		uint32 val = read_memory_mapped_io8(addr);
		val |= read_memory_mapped_io8(addr + 1) << 8;
		return val;
	}
	virtual void write_memory_mapped_io32(uint32 addr, uint32 data) {
		write_memory_mapped_io8(addr, data & 0xff);
		write_memory_mapped_io8(addr + 1, (data >> 8) & 0xff);
		write_memory_mapped_io8(addr + 2, (data >> 16) & 0xff);
		write_memory_mapped_io8(addr + 3, (data >> 24) & 0xff);
	}
	virtual uint32 read_memory_mapped_io32(uint32 addr) {
		uint32 val = read_memory_mapped_io8(addr);
		val |= read_memory_mapped_io8(addr + 1) << 8;
		val |= read_memory_mapped_io8(addr + 2) << 16;
		val |= read_memory_mapped_io8(addr + 3) << 24;
		return val;
	}
	virtual void write_memory_mapped_io8w(uint32 addr, uint32 data, int* wait) {
		*wait = 0;
		write_memory_mapped_io8(addr, data);
	}
	virtual uint32 read_memory_mapped_io8w(uint32 addr, int* wait) {
		*wait = 0;
		return read_memory_mapped_io8(addr);
	}
	virtual void write_memory_mapped_io16w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1;
		write_memory_mapped_io8w(addr, data & 0xff, &wait0);
		write_memory_mapped_io8w(addr + 1, (data >> 8) & 0xff, &wait1);
		*wait = wait0 + wait1;
	}
	virtual uint32 read_memory_mapped_io16w(uint32 addr, int* wait) {
		int wait0, wait1;
		uint32 val = read_memory_mapped_io8w(addr, &wait0);
		val |= read_memory_mapped_io8w(addr + 1, &wait1) << 8;
		*wait = wait0 + wait1;
		return val;
	}
	virtual void write_memory_mapped_io32w(uint32 addr, uint32 data, int* wait) {
		int wait0, wait1, wait2, wait3;
		write_memory_mapped_io8w(addr, data & 0xff, &wait0);
		write_memory_mapped_io8w(addr + 1, (data >> 8) & 0xff, &wait1);
		write_memory_mapped_io8w(addr + 2, (data >> 16) & 0xff, &wait2);
		write_memory_mapped_io8w(addr + 3, (data >> 24) & 0xff, &wait3);
		*wait = wait0 + wait1 + wait2 + wait3;
	}
	virtual uint32 read_memory_mapped_io32w(uint32 addr, int* wait) {
		int wait0, wait1, wait2, wait3;
		uint32 val = read_memory_mapped_io8w(addr, &wait0);
		val |= read_memory_mapped_io8w(addr + 1, &wait1) << 8;
		val |= read_memory_mapped_io8w(addr + 2, &wait2) << 16;
		val |= read_memory_mapped_io8w(addr + 3, &wait3) << 24;
		*wait = wait0 + wait1 + wait2 + wait3;
		return val;
	}

	// device to device
	typedef struct {
		DEVICE *device;
		int id;
		uint32 mask;
		int shift;
	} output_t;

	typedef struct {
		int count;
		output_t item[MAX_OUTPUT];
	} outputs_t;

	virtual void init_output_signals(outputs_t *items) {
		items->count = 0;
	}
	virtual void register_output_signal(outputs_t *items, DEVICE *device, int id, uint32 mask, int shift) {
		int c = items->count++;
		items->item[c].device = device;
		items->item[c].id = id;
		items->item[c].mask = mask;
		items->item[c].shift = shift;
	}
	virtual void register_output_signal(outputs_t *items, DEVICE *device, int id, uint32 mask) {
		int c = items->count++;
		items->item[c].device = device;
		items->item[c].id = id;
		items->item[c].mask = mask;
		items->item[c].shift = 0;
	}
	virtual void write_signals(outputs_t *items, uint32 data) {
		for(int i = 0; i < items->count; i++) {
			output_t *item = &items->item[i];
			int shift = item->shift;
			uint32 val = (shift < 0) ? (data >> (-shift)) : (data << shift);
			uint32 mask = (shift < 0) ? (item->mask >> (-shift)) : (item->mask << shift);
			item->device->write_signal(item->id, val, mask);
		}
	};
	virtual void write_signal(int id, uint32 data, uint32 mask) {}
	virtual uint32 read_signal(int ch) {
		return 0;
	}

	// z80 daisy chain
	virtual void set_context_intr(DEVICE* device, uint32 bit) {}
	virtual void set_context_child(DEVICE* device) {}

	// interrupt device to device
	virtual void set_intr_iei(bool val) {}

	// interrupt device to cpu
	virtual void set_intr_line(bool line, bool pending, uint32 bit) {}

	// interrupt cpu to device
	virtual uint32 intr_ack() {
		return 0xff;
	}
	virtual void intr_reti() {}
	virtual void intr_ei() {}

	// dma
	virtual void do_dma() {}

	// cpu
	virtual int run(int clock) {
		// when clock == -1, run one opecode
		return 0;
	}
#ifdef USE_CPU_REAL_MACHINE_CYCLE
	virtual int run(int clock, int accum) {
		return 0;
	}
#endif
	virtual uint32 get_pc() {
		return 0;
	}

	// bios
	virtual bool bios_call(uint32 PC, uint16 regs[], uint16 sregs[], int32* ZeroFlag, int32* CarryFlag) {
		return false;
	}
	virtual bool bios_int(int intnum, uint16 regs[], uint16 sregs[], int32* ZeroFlag, int32* CarryFlag) {
		return false;
	}

	// event manager
	DEVICE* event_manager;

	virtual void set_context_event_manager(DEVICE* device) {
		event_manager = device;
	}
	virtual int event_manager_id() {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		return event_manager->this_device_id;
	}
	virtual void register_event(DEVICE* device, int event_id, double usec, bool loop, int* register_id) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->register_event(device, event_id, usec, loop, register_id);
	}
	virtual void register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->register_event_by_clock(device, event_id, clock, loop, register_id);
	}
	virtual void cancel_event(int register_id) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->cancel_event(register_id);
	}
	virtual void register_frame_event(DEVICE* device) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->register_frame_event(device);
	}
	virtual void register_vline_event(DEVICE* device) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->register_vline_event(device);
	}
	virtual uint64 current_clock() {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		return event_manager->current_clock();
	}
	virtual uint64 passed_clock(uint64 prev) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		return event_manager->passed_clock(prev);
	}
	virtual uint32 get_cpu_pc(int index) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		return event_manager->get_cpu_pc(index);
	}
	virtual void set_frames_per_sec(double frames) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->set_frames_per_sec(frames);
	}
	virtual void set_lines_per_frame(int lines) {
		if(event_manager == NULL) {
			event_manager = vm->first_device->next_device;
		}
		event_manager->set_lines_per_frame(lines);
	}
	virtual void update_timing(int new_clocks, double new_frames_per_sec, int new_lines_per_frame) {}

	// event callback
	virtual void event_callback(int event_id, int err) {}
	virtual void event_pre_frame() {}	// this event is to update timing settings
	virtual void event_frame() {}
	virtual void event_vline(int v, int clock) {}
	virtual void event_hsync(int v, int h, int clock) {}

	// sound
	virtual void mix(int32* buffer, int cnt) {}
	virtual void set_volume(int volume) {}
	virtual void set_volume(int volumel, int volumer) {}

	DEVICE* prev_device;
	DEVICE* next_device;
	int this_device_id;

	char this_class_name[13];
	char this_identifier[5];

#ifdef USE_EMU_INHERENT_SPEC
	// display
	virtual void update_vram(int v, int clock) {}

	bool now_reset;
#endif
	//for resume
#pragma pack(1)
	typedef struct vm_state_ident_st {
		char   class_name[12];
		char   identifier[4];
		uint32 size;
		uint16 version;
		int    device_id;
		char   reserved[6];
	} vm_state_ident_t;
	vm_state_ident_t vm_state_ident;
#pragma pack()

	bool find_state_chunk(FILEIO *fio, char *class_name, char *identifier, vm_state_ident_t *vm_state_i) {
		uint32 size = sizeof(vm_state_ident_t);

		for(int retry = 0; retry < 2; retry++) {
			while (fio->Fread(vm_state_i, size, 1) == 1) {
				if (strncmp(class_name, vm_state_i->class_name, 12) == 0
				&&  strncmp(identifier, vm_state_i->identifier, 4) == 0) {
					// match chunk
					return true;
				}
				fio->Fseek(vm_state_i->size - size, FILEIO_SEEK_CUR);
			}
			// retry first
			fio->Fseek(sizeof(vm_state_header_t), FILEIO_SEEK_SET);
		}
		return false;
	}
	bool find_state_chunk(FILEIO *fio, vm_state_ident_t *vm_state_i) {
		return find_state_chunk(fio, this_class_name, this_identifier, vm_state_i);
	}

#define FIND_STATE_CHUNK(p_fio, st_i) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		p_fio->Fseek(-(long)sizeof(st_i), FILEIO_SEEK_CUR); \
	}

#define READ_STATE_CHUNK(p_fio, st_i, st_v) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		memset(&st_v, 0, sizeof(st_v)); \
		if (st_i.size >= (sizeof(st_v) + sizeof(st_i))) { \
			p_fio->Fread(&st_v, sizeof(st_v), 1); \
			p_fio->Fseek(st_i.size - sizeof(st_v) - sizeof(st_i), FILEIO_SEEK_CUR); \
		} else { \
			p_fio->Fread(&st_v, st_i.size - sizeof(st_i), 1); \
		} \
	}

#define READ_STATE_VARIABLE_CHUNK(p_fio, st_i, st_v) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		memset(&st_v, 0, sizeof(st_v)); \
		if (st_i.size >= (sizeof(st_v) + sizeof(st_i))) { \
			p_fio->Fread(&st_v, sizeof(st_v), 1); \
		} else { \
			p_fio->Fread(&st_v, st_i.size - sizeof(st_i), 1); \
		} \
	}

	// for debug
	virtual void dummyf(const void *format, ...) {}
};

#endif
